#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2015, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

"""Subcommands are now implemented using a modular design that allows extending
jug with additional commands.
This API is currently in experimental stage and may change in the future.

The following serves as an example of how to extend jug's commands.

Lets assume you wanted to create a custom report and have it available as::

    $ jug my-fancy-report

One way to achieve this is to add the following code to ``~/.config/jug/jug_user_commands.py``::

    from jug.subcommands import SubCommand

    class FancyReport(SubCommand):
        "Produces a fancy report of my results"

        name = "my-fancy-report"

        def run(self, *args, **kwargs):
            ...

    fancy_report = FancyReport()

The first line of the class docstring is important as it will be shown in jug's
usage help page. The name attribute is also required and should be the name of
your subcommand on the command-line.

The body of the method ``run()`` defines what should happen when you
call the subcommand ``jug my-fancy-report``.

The ``run`` function will receive the following objects::

* ``options``  - object representing command-line and user options
* ``store``    - backend object reponsible for handling jobs
* ``jugspace`` - a namespace of jug internal variables (better not touch)

additional objects may be introduced in the future so make sure your function
uses ``*args, **kwargs`` to maintain compatibility.

Finally, in order to register the subcommand, you must instanciate the subcommand.


If your subcommand needs configurable options you can expose them via command-line
by defining two additional methods::

    class FancyReport(SubCommand):
        ...

        def parse(self, parser):
            parser.add_argument('--tofile', action='store',
                                dest='report_tofile',
                                help='Name of file to use for report')

        def parse_defaults(self):
            return {
                "report_tofile": "report.txt",
            }

    fancy_report = FancyReport()

The first method configures argparse arguments that will be available as
``jug my-fancy-report --tofile myreport.txt``. These will also be avaiable to
the ``run()`` method as part of the ``options`` object.

The second defines default values in case of omission. The ``key`` should match
the ``dest=`` attribute of ``add_argument()`` and the ``value`` should be any object
to be used by your ``run()`` method. Note that the value received in the command-line
will be automatically converted to the same type as this default (i.e. if your default
is ``True`` any ``--tofile john`` would result in ``bool("john") -> True``).

For more information on parser configuration refer to ``argparse``'s documentation.

NOTE: A few words of caution, we cannot rely on ``argparse``'s ``default=`` option since
it doesn't allow distinguishing between user supplied and built-in (default) values.
For the same reason, don't use ``action=`` with ``store_true`` or ``store_false``
instead use ``store_const`` and ``const=`` with ``True`` or ``False``.
Failing to do so will cause any matching setting on ``jugrc`` to not have any effect.
"""

__all__ = [
    'cmdapi',
    'SubCommand',
    'SubCommandError',
    'NoSuchCommandError',
]

import importlib
import logging
import os
import pkgutil
import sys
import traceback
from ..options import Options
from ..jug_version import CITATION
from abc import ABCMeta, abstractmethod, abstractproperty
from six import add_metaclass


class SubCommandError(Exception):
    "Exception raised when a subcommand doesn't respect the API"


class NoSuchCommandError(Exception):
    "Exception raised when a subcommand doesn't respect the API"


def _get_helptext(command):
    "First line of the docstring is to be shown on the help/usage text"
    try:
        return command.__doc__.splitlines()[0]
    except AttributeError:
        raise SubCommandError("Command '%s' is missing a documentation string" % (command,))


def _invalid_module(module, e):
    raise SubCommandError(
        """Invalid subcommand structure.
Please make sure that the subcommand(s) '%s' conform(s) to the API.

    help(jug.subcommands) for more information

Original error was:
%s
""" % (module, e))


@add_metaclass(ABCMeta)
class SubCommand:
    """Define a subcommand and its command-line options
    """

    def __init__(self):
        cmdapi._register(self.name, self)
        cmdapi.update_defaults(self.name)

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    @abstractproperty
    def name(self):
        pass

    def parse(self, parser):
        """Define command line options using parser.add_argument()

        The parser object is an argparser subparser group.
        Anything returned by this method is ignored
        """
        pass

    def parse_defaults(self):
        """Define default values for parser options

        Should return a dictionary mapping ``dest=`` targets to their default.
        """
        pass

    @abstractmethod
    def run(self, *args, **kwargs):
        """Re-define to specify what your subcommand is supposed to do

        This code will receive the following arguments:
        * ``options``  - object representing command-line and user options
        * ``store``    - backend object reponsible for handling jobs
        * ``jugspace`` - a namespace of jug internal variables (better not touch)

        Anything returned by this method is ignored
        """
        pass


class SubCommandDict(dict):
    def __getitem__(self, command):
        if command not in self:
            self.load_commands(command)

        return super(SubCommandDict, self).__getitem__(command)

    def load_commands(self, stop_on_command=None):
        """Load all modules in jug's subcommands and user's jug folder

        If stop_on_command is given the function will return True as soon as
        a matching command is found
        """
        for _, name, _ in pkgutil.iter_modules(__path__):
            module = __name__ + '.' + name

            self._try_import(module)

            if stop_on_command in self:
                return True

        self._load_user_commands()

    def _try_import(self, module):
        try:
            importlib.import_module(module)
        except Exception as e:
            logging.warning("Couldn't load subcommand '%s' with error '%s'", module, e)

    def _load_user_commands(self, user_path="~/.config/jug/"):
        user_path = os.path.expanduser(user_path)
        logging.debug("Loading user commands from '%s'", user_path)
        if os.path.isdir(user_path):
            if user_path not in sys.path:
                logging.debug("Adding path '%s' to PYTHONPATH", user_path)
                sys.path.insert(0, user_path)

            user_commands = os.path.join(user_path, "jug_user_commands.py")
            if os.path.isfile(user_commands):
                self._try_import("jug_user_commands")


class SubCommandManager:
    def __init__(self):
        self._commands = SubCommandDict()
        self.default_options = Options(None)

    def _register(self, name, cmd_instance):
        if name in self._commands and self._commands[name] != cmd_instance:
            logging.warning("Jug: command: '%s' will be overriden with code from '%s'",
                            name, cmd_instance.__class__.__name__)

        self._commands[name] = cmd_instance

    def update_defaults(self, name):
        cmd = self.get(name)
        opts = cmd.parse_defaults()
        if opts is not None:
            self.default_options.update(opts)

    def get(self, command):
        try:
            return self._commands[command]
        except KeyError:
            raise NoSuchCommandError("Unknown subcommand '%s'" % (command,))

    def run(self, command, *args, **kwargs):
        """Execute subcommand
        """
        try:
            cmd = self.get(command)
        except NoSuchCommandError as e:
            self.usage(error=e)

        try:
            return cmd(*args, **kwargs)
        except TypeError as e:
            if "unexpected keyword" in str(e):
                _invalid_module(command, e)

            traceback.print_exc(file=sys.stderr)

            self.usage(error=e)

    def usage(self, error='', exit=True, _print=True, *args, **kwargs):
        "Shows help/usage information"

        usage_text = ['''\
jug SUBCOMMAND [JUGFILE] [OPTIONS...]

Docs: https://jug.readthedocs.io/
Copyright: 2008-2017, Luis Pedro Coelho
Citation: http://doi.org/10.5334/jors.161

If you use Jug for generating results for a peer-reviewed publication, please
cite:

    Coelho, L.P., (2017). Jug: Software for Parallel Reproducible Computation in
    Python. Journal of Open Research Software. 5(1), p.30.

    http://doi.org/10.5334/jors.161


Subcommands
-----------
''']
        self._commands.load_commands()

        for name, cmd in sorted(self._commands.items()):
            usage_text.append("   %-15s %s" % (name + ":", _get_helptext(cmd)))

        usage_text.append("\nhelp:")
        usage_text.append("  Use 'jug <subcommand> --help' for subcommand specific options")

        if error:
            usage_text.append("")
            usage_text.append(str(error))

        message = "\n".join(usage_text) + "\n \n"

        if _print:
            sys.stdout.write(message)

        if exit:
            sys.exit(1)

        return message

    def get_subcommand_parsers(self, subparsers):
        self._commands.load_commands()

        parsers = []

        for name, cmd in sorted(self._commands.items()):
            parser = subparsers.add_parser(
                name,
                # This is necessary to have all the same output on all subparsers
                usage=self.usage(_print=False, exit=False),
            )
            parsers.append(parser)

            group = parser.add_argument_group(name)
            cmd.parse(group)

        return parsers

def maybe_print_citation_info(options):
    '''Unless options.will_cite, prints citation information'''
    if not options.will_cite:
        print("If you use Jug in a published research paper please cite")
        print(CITATION)
        print('')
        print('Use the --will-cite option to suppress this message')
        print('(Or set will_cite = True in the Jug configuration file)')
        print('')

cmdapi = SubCommandManager()



#-*- coding: utf-8 -*-
# Copyright (C) 2009-2016, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

'''
redis_store: store based on a redis backend
'''


import re
import six
import logging
from base64 import b64encode, b64decode
from jug.backends.encode import encode, decode
from .base import base_store, base_lock


try:
    import redis
    redis_functional = True
except ImportError:
    redis = None
    redis_functional = False

def _resultname(name):
    if type(name) == type(six.u('')):
        name = name.encode('utf-8')
    return six.b('result:') + name

def _lockname(name):
    if type(name) == type(six.u('')):
        name = name.encode('utf-8')
    return six.b('lock:') + name

_LOCKED = 1

_redis_urlpat = re.compile(r'redis://(?P<host>[A-Za-z0-9\.\-]+)(\:(?P<port>[0-9]+))?/?')


class redis_store(base_store):
    def __init__(self, url):
        '''
        '''
        if redis is None:
            raise IOError('jug.redis_store: redis module is not found!')
        redis_params = {}
        match = _redis_urlpat.match(url)
        if match:
            redis_params = match.groupdict()
            if redis_params['port'] == None:
                del redis_params['port']
            else:
                redis_params['port'] = int( redis_params['port'] )
        logging.info('connecting to %s' % redis_params)

        self.redis = redis.Redis(**redis_params)


    def dump(self, object, name):
        '''
        dump(object, name)
        '''
        s = encode(object)
        if s:
            s = b64encode(s)
        self.redis.set(_resultname(name), s)


    def can_load(self, name):
        '''
        can = can_load(name)
        '''
        return self.redis.exists(_resultname(name))


    def load(self, name):
        '''
        obj = load(name)

        Loads the object identified by `name`.
        '''
        s = self.redis.get(_resultname(name))
        if s:
            s = b64decode(s)
        return decode(s)


    def remove(self, name):
        '''
        was_removed = remove(name)

        Remove the entry associated with name.

        Returns whether any entry was actually removed.
        '''
        return self.redis.delete(_resultname(name))


    def cleanup(self, active, keeplocks=False):
        '''
        cleanup()

        Implement 'cleanup' command
        '''
        existing = list(self.list())
        for act in active:
            try:
                existing.remove(_resultname(act.hash()))
            except ValueError:
                pass

        if keeplocks:
            for lock in self.listlocks():
                try:
                    existing.remove(_lockname(lock))
                except ValueError:
                    pass

        cleaned = len(existing)
        for superflous in existing:
            self.redis.delete(_resultname(superflous))

        return cleaned

    def remove_locks(self):
        locks = self.redis.keys('lock:*')
        for lk in locks:
            self.redis.delete(lk)
        return len(locks)

    def list(self):
        existing = self.redis.keys('result:*')
        for ex in existing:
            yield ex[len('result:'):]

    def listlocks(self):
        locks = self.redis.keys('lock:*')
        for lk in locks:
            yield lk[len('lock:'):]


    def getlock(self, name):
        return redis_lock(self.redis, name)


    def close(self):
        # It seems some versions of the protocol are implemented differently
        # and do not have the ``disconnect`` method
        try:
            self.redis.disconnect()
        except:
            pass



class redis_lock(base_lock):
    '''
    redis_lock

    Functions:
    ----------

        * get(): acquire the lock
        * release(): release the lock
        * is_locked(): check lock state
    '''

    def __init__(self, redis, name):
        self.name = _lockname(name)
        self.redis = redis


    def get(self):
        '''
        lock.get()
        '''
        previous = self.redis.getset(self.name, _LOCKED)
        return (previous is None)


    def release(self):
        '''
        lock.release()

        Removes lock
        '''
        self.redis.delete(self.name)


    def is_locked(self):
        '''
        locked = lock.is_locked()
        '''
        status = self.redis.get(self.name)
        return status is not None and status == _LOCKED




from jug import TaskGenerator
import jug.mapreduce
import math

@TaskGenerator
def double(x):
    return x*2

@TaskGenerator
def sum2(a,b):
    return (a+b)

sumtwo = jug.mapreduce.mapreduce(sum2, double, list(range(10)))



from nose.tools import with_setup, raises

from jug.hooks import register_hook, register_hook_once, reset_all_hooks, jug_hook

A = []
def inc_A():
    A.append(0)

def teardown():
    reset_all_hooks()
    del A[:] # A.clear() is not available in Python 2.x

@with_setup(teardown=teardown)
def test_basic():
    assert len(A) == 0
    register_hook('execute.task-executed1', inc_A)
    jug_hook('execute.task-executed1')
    assert len(A) == 1

@with_setup(teardown=teardown)
def test_register_twice():
    register_hook('execute.task-executed1', inc_A)
    register_hook('execute.task-executed1', inc_A)
    jug_hook('execute.task-executed1')
    assert len(A) == 2

@with_setup(teardown=teardown)
def test_reset_all_hooks():
    register_hook('execute.task-executed1', inc_A)
    jug_hook('execute.task-executed1')
    assert len(A) == 1
    reset_all_hooks()
    jug_hook('execute.task-executed1')
    assert len(A) == 1

@with_setup(teardown=teardown)
def test_register_once():
    assert len(A) == 0
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    register_hook_once('execute.task-executed1', 'just-testing-mate', inc_A)
    assert len(A) == 0
    jug_hook('execute.task-executed1')
    assert len(A) == 1

@raises(ValueError)
def test_bad_name():
    register_hook('my_bad_hook', inc_A)




import inspect
import os

from jug.subcommands import status
from jug.tests.task_reset import task_reset
from jug.tests.utils import simple_execute
from jug.options import default_options
import jug


_jugdir = os.path.abspath(inspect.getfile(inspect.currentframe()))
_jugdir = os.path.join(os.path.dirname(_jugdir), 'jugfiles')


@task_reset
def test_nocache():
    jugfile = os.path.join(_jugdir, 'simple.py')
    store, space = jug.jug.init(jugfile, 'dict_store')
    simple_execute()

    options = default_options.copy()
    options.jugdir = store
    options.jugfile = jugfile
    options.verbose = 'quiet'
    assert status.status(options) == 8 * 4

@task_reset
def test_cache():
    jugfile = os.path.join(_jugdir, 'simple.py')
    store, space = jug.jug.init(jugfile, 'dict_store')

    options = default_options.copy()
    options.jugdir = store
    options.jugfile = jugfile
    options.verbose = 'quiet'
    options.status_cache = True
    options.status_cache_file = ':memory:'

    assert status.status(options) == 0
    simple_execute()
    assert status.status(options) == 8 * 4

@task_reset
def test_cache_bvalue():
    jugfile = os.path.join(_jugdir, 'bvalue.py')
    store, space = jug.jug.init(jugfile, 'dict_store')

    options = default_options.copy()
    options.jugdir = store
    options.jugfile = jugfile
    options.verbose = 'quiet'
    options.status_cache = True
    options.status_cache_file = ':memory:'

    assert status.status(options) == 0
    simple_execute()
    assert status.status(options) == 1




# -*- coding: utf-8 -*-
# Copyright (C) 2008-2013, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
# License : MIT

'''
mapreduce: Build tasks that follow a map-reduce pattern.
'''


from .task import Task
from .utils import identity
from .hash import hash_one

from itertools import chain

__all__ = [
    'mapreduce',
    'map',
    'reduce',
    ]

def _get_function(f):
    if getattr(f, '_jug_is_task_generator', False):
        hvalue = hash_one(f)
        f = f.f
        f.__jug_hash__ = lambda: hvalue
        return f
    return f

def _jug_map_reduce(reducer, mapper, inputs):
    from six.moves import reduce
    reducer = _get_function(reducer)
    mapper = _get_function(mapper)
    return reduce(reducer, _jug_map(mapper, inputs))

def _jug_reduce(reducer, inputs):
    from six.moves import reduce
    reducer = _get_function(reducer)
    return reduce(reducer, chain(inputs))

def _break_up(lst, step):
    start = 0
    next = step
    while start < len(lst):
        yield lst[start:next]
        start = next
        next += step


def _jug_map(mapper, es):
    if mapper is None:
        mapper = lambda x: x
    return [mapper(e) for e in es]

def _jug_map_curry(mapper, es):
    return [mapper(*e) for e in es]


def mapreduce(reducer, mapper, inputs, map_step=4, reduce_step=8):
    '''
    task = mapreduce(reducer, mapper, inputs, map_step=4, reduce_step=8)

    Create a task that does roughly the following::

        reduce(reducer, map(mapper, inputs))

    Roughly because the order of operations might be different. In particular,
    `reducer` should be a true `reducer` functions (i.e., commutative and
    associative).

    Parameters
    ----------
    reducer : associative, commutative function
            This should map
                  Y_0,Y_1 -> Y'
    mapper : function from X -> Y
    inputs : list of X

    map_step : integer, optional
            Number of mapping operations to do in one go.
            This is what defines an inner task. (default: 4)
    reduce_step : integer, optional
            Number of reduce operations to do in one go.
            (default: 8)

    Returns
    -------
    task : jug.Task object
    '''
    reducers = [Task(_jug_map_reduce, reducer, mapper, input_i) for input_i in _break_up(inputs, map_step)]
    while len(reducers) > 1:
        reducers = [Task(_jug_reduce, reducer, reduce_i) for reduce_i in _break_up(reducers, reduce_step)]
    if len(reducers) == 0:
        return identity([])
    elif len(reducers) == 1:
        return reducers[0]
    else:
        assert False, 'This is a bug'

class block_access_slice(object):
    __slots__ = ('base', 'start', 'stop', 'stride', '_hvalue')
    def __init__(self, access, orig):
        self.base = access
        self.start,self.stop,self.stride = orig
        self._hvalue = None

    def __getitem__(self, p):
        if isinstance(p, slice):
            start,stop,stride = p.indices(len(self))
            return block_access_slice(self.base, (self.start + start, self.stop - (len(self)-stop), self.stride * stride))
        elif isinstance(p, int):
            p *= self.stride
            p += self.start
            if p >= self.stop:
                raise IndexError
            return self.base[p]
        else:
            raise TypeError

    def __len__(self):
        return self.stop - self.start

    def __jug_hash__(self):
        if self._hvalue is not None:
            return self._hvalue
        self._hvalue = hash_one({
            'type': 'map-access-slice',
            'base': self.base,
            'start': self.start,
            'stop': self.stop,
            'stride': self.stride,
        })
        return self._hvalue

    def __jug_value__(self):
        from .task import value
        return [value(self[i]) for i in range(len(self))]

class block_access(object):
    __slots__ = ('blocks','block_size', 'len','_hvalue')
    def __init__(self, blocks, block_size, len):
        self.blocks = blocks
        self.block_size = block_size
        self.len = len
        self._hvalue = None

    def __getitem__(self, p):
        if isinstance(p, slice):
            return block_access_slice(self, p.indices(self.len))
        elif isinstance(p, int):
            if not (0 <= p < self.len):
                raise IndexError
            b = p//self.block_size
            bi = p % self.block_size
            return self.blocks[b][bi]
        else:
            raise TypeError

    def __len__(self):
        return self.len

    def __jug_hash__(self):
        if self._hvalue is not None:
            return self._hvalue
        value = hash_one({
            'type': 'map-access',
            'len': self.len,
            'blocks': self.blocks,
            'block_size': self.block_size,
        })
        self._hvalue = value
        return value

    def __jug_value__(self):
        from .task import value
        return [value(self[i]) for i in range(len(self))]

def map(mapper, sequence, map_step=4):
    '''
    sequence' = map(mapper, sequence, map_step=4)

    Roughly equivalent to::

        sequence' = [Task(mapper,s) for s in sequence]

    except that the tasks are grouped in groups of `map_step`

    Parameters
    ----------
    mapper : function
        function from A -> B
    sequence : list of A
    map_step : integer, optional
        nr of elements to process per task. This should be set so that each
        task takes the right amount of time.

    Returns
    -------
    sequence' : list of B
        sequence'[i] = mapper(sequence[i])

    See Also
    --------
    mapreduce
    currymap: function
        Curried version of this function
    '''
    if map_step == 1:
        return [Task(mapper, s) for s in sequence]
    blocks = []
    n = 0
    for ss in _break_up(sequence, map_step):
        blocks.append(
            Task(_jug_map, _get_function(mapper), ss)
            )
        n += len(ss)
    return block_access(blocks, map_step, n)

def currymap(mapper, sequence, map_step=4):
    '''
    sequence' = currymap(mapper, sequence, map_step=4)

    Roughly equivalent to::

        sequence' = [Task(mapper,*s) for s in sequence]

    except that the tasks are grouped in groups of `map_step`

    Parameters
    ----------
    mapper : function
        function from A1 -> A2 ... -> An -> B
    sequence : list of (A1,A2,...,An)
    map_step : integer, optional
        nr of elements to process per task. This should be set so that each
        task takes the right amount of time.

    Returns
    -------
    sequence' : list of B
        sequence'[i] = mapper(*sequence[i])

    See Also
    --------
    mapreduce: function
    map: function
        Uncurried version of this function
    '''
    if map_step == 1:
        return [Task(mapper, *s) for s in sequence]
    result = []
    for ss in _break_up(sequence, map_step):
        t = Task(_jug_map_curry, _get_function(mapper), ss)
        for i,_ in enumerate(ss):
            result.append(t[i])
    return result




def reduce(reducer, inputs, reduce_step=8):
    '''
    task = reduce(reducer, inputs, reduce_step=8)

    Parameters
    ----------
    reducer : associative, commutative function
            This should map
                  Y_0,Y_1 -> Y'
    inputs : list of X
    reduce_step : integer, optional
            Number of reduce operations to do in one go.
            (default: 8)

    Returns
    -------
    task : jug.Task object

    See Also
    --------
    mapreduce
    '''
    return mapreduce(reducer, None, inputs, reduce_step=reduce_step)




import jug.options

from six import StringIO
from nose.tools import raises

def test_chaining():
    first = jug.options.Options(None)
    second = jug.options.Options(first)
    third = jug.options.Options(second)

    first.one = 'one'
    second.two = 'two'
    third.three = 'three'

    assert third.one == 'one'
    first.one = 1
    assert third.one == 1
    assert second.one == 1
    assert second.two == 'two'
    assert third.three == 'three'
    @raises(AttributeError)
    def not_present_key(obj, key):
        return obj.key

    yield not_present_key, first, 'two'
    yield not_present_key, second, 'three'

_options_file = '''
[main]
jugfile=myjugfile.py

[execute]
wait-cycle-time=23
'''

def test_parse():
    parsed = jug.options.parse(
        ["execute", "--pdb"],
        StringIO(_options_file))

    assert parsed.jugfile == 'myjugfile.py'
    assert parsed.pdb
    assert parsed.execute_wait_cycle_time == 23
    assert not parsed.aggressive_unload

def test_copy():
    parsed = jug.options.parse(
        ["execute", "--pdb"],
        StringIO(_options_file))
    copy = parsed.copy()
    assert parsed.pdb
    assert not parsed.aggressive_unload

def test_sys_argv():
    from sys import argv
    parsed = jug.options.parse(
        ["execute", "--pdb"])
    assert argv == ['jugfile.py']

    parsed = jug.options.parse(
        ["execute", 'jugfile.py'])
    assert argv == ['jugfile.py']

    parsed = jug.options.parse(
        ["execute", 'jugfile.py', '--', 'arg'])
    assert argv == ['jugfile.py', 'arg']

    parsed = jug.options.parse(
        ["execute", 'jugfile.py', 'arg'])
    assert argv == ['jugfile.py', 'arg']

    parsed = jug.options.parse(
        ["execute", 'jugfile.py', 'arg', '--', '--hello'])
    assert argv == ['jugfile.py', 'arg', '--hello']

def test_bool():
    from jug.options import _str_to_bool
    assert not _str_to_bool("")
    assert not _str_to_bool("0")
    assert not _str_to_bool("false")
    assert not _str_to_bool("FALSE")
    assert not _str_to_bool("off")
    assert _str_to_bool("on")
    assert _str_to_bool("true")
    assert _str_to_bool("1")



from sys import exit
def exit_if_file_exists(fname):
    '''Before each task execute, check if file exists. If so, exit.

    Note that a check is only performed before a Task is execute. Thus, jug
    will not exit immediately if it is executing another long-running task.

    Parameters
    ----------
    fname : str
        path to check
    '''
    from jug import hooks
    def check_file(_t):
        from os import path
        if path.exists(fname):
            exit(0)
    hooks.register_hook('execute.task-pre-execute', check_file)

def exit_when_true(f, function_takes_Task=False):
    '''Generic exit check.
    
    After each task, call function ``f`` and exit if it return true.

    Parameters
    ----------
    f : function
        Function to call
    function_takes_Task : boolean, optional
        Whether to call the function with the task just executed (default: False)
    '''

    from jug import hooks
    if not function_takes_Task:
        f = lambda t : f()
    def exit_when(t):
        if f(t):
           exit(0)
    hooks.register_hook('execute.task-executed1', exit_when)

def exit_after_n_tasks(n):
    '''Exit after a specific number of tasks have been executed
    
    Parameters
    ----------
    n : int
        Number of tasks to execute
    '''
    from jug import hooks
    # In newer Python, we could use nonlocal, but this is a work around
    # (http://stackoverflow.com/questions/9603278/is-there-something-like-nonlocal-in-python-3/9603491#9603491)
    executed = [0]
    def exit_after(_t):
        executed[0] += 1
        if executed[0] >= n:
           exit(0)
    hooks.register_hook('execute.task-executed1', exit_after)

def exit_after_time(hours=0, minutes=0, seconds=0):
    '''Exit after a specific number of tasks have been executed

    Note that this only checks the time **after each task has finished
    executing**. Thus if you are using this to limit the amount of time each
    process takes, make sure to specify a lower limit than what is needed.

    Parameters
    ----------
    hours : number, optional
    minutes : number, optional
    seconds : number, optional
    '''
    from jug import hooks
    from time import time
    deadline = time()
    deadline += seconds
    deadline += 60*minutes
    deadline += 60*60*hours

    def check_time(_t):
        if time() >= deadline:
            exit(0)
    hooks.register_hook('execute.task-executed1', check_time)

def exit_env_vars(environ=None):
    '''
    exit_env_vars(environ={os.environ})

    Set exit markers based on the environment.

    The following variables are used if they are set (if they are not set, they
    are ignored).

    JUG_MAX_TASKS: Maximum nr. of tasks.

    JUG_MAX_HOURS: Maximum hours
    JUG_MAX_MINUTES: Maximum minutes
    JUG_MAX_SECONDS: Maximum seconds

    For the time based limits, see the comment on `exit_after_time` on how
    these limits are not strict as they are only checked after each task
    completion event.

    If either of the variables above is set, its value should be an int or an
    error will be raised.

    JUG_EXIT_IF_FILE_EXISTS: Set exit file name

    See Also
    --------
    exit_after_n_tasks
    exit_after_time
    exit_if_file_exists
    '''

    if environ is None:
        import os
        environ = os.environ
    if 'JUG_MAX_TASKS' in environ:
        exit_after_n_tasks(int(environ['JUG_MAX_TASKS']))
    
    hours = 0
    minutes = 0
    seconds = 0
    if 'JUG_MAX_HOURS' in environ:
        hours = int(environ['JUG_MAX_HOURS'])
    if 'JUG_MAX_MINUTES' in environ:
        minutes = int(environ['JUG_MAX_MINUTES'])
    if 'JUG_MAX_SECONDS' in environ:
        seconds = int(environ['JUG_MAX_SECONDS'])
    if hours or minutes or seconds:
        exit_after_time(hours=hours, minutes=minutes, seconds=seconds)

    if 'JUG_EXIT_IF_FILE_EXISTS' in environ:
        exit_if_file_exists(os['JUG_EXIT_IF_FILE_EXISTS'])



# -*- coding: utf-8 -*-
# Copyright (C) 2008-2016, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.


from sys import exit
try:
    import setuptools
except:
    from sys import stdout
    stdout.write('''
setuptools not found. Please install it.

On linux, the package is often called python-setuptools\n''')
    exit(1)

exec(compile(open('jug/jug_version.py').read(), 'jug/jug_version.py', 'exec'))
long_description = open('README.rst').read()

classifiers = [
'Development Status :: 5 - Production/Stable',
'Environment :: Console',
'License :: OSI Approved :: MIT License',
'Operating System :: POSIX',
'Operating System :: OS Independent',
'Programming Language :: Python',
'Programming Language :: Python :: 2',
'Programming Language :: Python :: 2.7',
'Programming Language :: Python :: 3',
'Programming Language :: Python :: 3.3',
'Programming Language :: Python :: 3.4',
'Programming Language :: Python :: 3.5',
'Programming Language :: Python :: 3.6',
'Topic :: Scientific/Engineering',
'Topic :: Software Development',
'Topic :: System :: Distributed Computing',
'Intended Audience :: Science/Research',
]

setuptools.setup(name = 'Jug',
      version = __version__,
      description = 'A Task Based Parallelization Framework',
      long_description = long_description,
      author = 'Luis Pedro Coelho',
      author_email = 'luis@luispedro.org',
      license = 'MIT',
      platforms = ['Any'],
      classifiers = classifiers,
      url = 'https://jug.readthedocs.io',
      packages = setuptools.find_packages(),
      entry_points={
          'console_scripts': [
              'jug = jug.jug:main',
          ],
      },
      scripts = ['bin/jug-execute'],
      test_suite = 'nose.collector',
      install_requires=['six', 'redis'],
      )





from jug import barrier, value, TaskGenerator
from jug.compound import CompoundTaskGenerator

@TaskGenerator
def double(x):
    return 2*x

@CompoundTaskGenerator
def twice(x):
    x2 = double(x)
    barrier()
    return double(value(x2))

four = double(2)
sixteen = twice(four)




#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2015, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from collections import defaultdict
from contextlib import contextmanager

import jug
from .. import task
from .. import backends
from ..task import Task
from ..backends import memoize_store
from ..io import print_task_summary_table
from . import SubCommand

__all__ = [
    'status'
]


unknown = 'unknown'
waiting = 'waiting'
ready = 'ready'
running = 'running'
finished = 'finished'


def create_sqlite3(connection, ht, deps, rdeps):
    connection.executescript('''
    CREATE TABLE ht (
            id INTEGER PRIMARY KEY,
            name CHAR(128),
            hash CHAR(128),
            status INT);
    CREATE TABLE dep (
        source INT,
        target INT);
    ''')

    connection.executemany('INSERT INTO ht VALUES(?,?,?,?)', ht)

    for i, cdeps in deps.items():
        if len(cdeps):
            connection.executemany('''
               INSERT INTO dep(source, target) VALUES(?,?)
                ''', [(i, cd) for cd in cdeps])


def retrieve_sqlite3(connection):
    ht = connection. \
            execute('SELECT * FROM ht ORDER BY id'). \
            fetchall()
    deps = defaultdict(list)
    rdeps = defaultdict(list)
    for d0, d1 in connection.execute('SELECT * FROM dep'):
        deps[d0].append(d1)
        rdeps[d1].append(d0)
    return ht, dict(deps), dict(rdeps)


def save_dirty3(connection, dirty):
    connection.executemany('UPDATE ht SET STATUS = ? WHERE id = ?', [(nstatus, id) for id, nstatus in dirty.items()])


@contextmanager
def _open_connection(options):
    import sqlite3
    connection = sqlite3.connect(options.status_cache_file)
    yield connection
    connection.commit()
    connection.close()


def load_jugfile(options):
    store, _ = jug.init(options.jugfile, options.jugdir)
    h2idx = {}
    ht = []
    deps = {}
    for i, t in enumerate(task.alltasks):
        deps[i] = [h2idx[d.hash() if isinstance(d, Task) else d._base_hash()]
                   for d in t.dependencies()]
        hash = t.hash()
        ht.append((i, t.name, hash, unknown))
        h2idx[hash] = i

    rdeps = defaultdict(list)
    for k, v in deps.items():
        for rv in v:
            rdeps[rv].append(k)
    return store, ht, deps, dict(rdeps)


def update_status(store, ht, deps, rdeps):
    tasks_waiting = defaultdict(int)
    tasks_ready = defaultdict(int)
    tasks_running = defaultdict(int)
    tasks_finished = defaultdict(int)

    store = memoize_store(store, list_base=True)
    dirty = {}
    for i, name, hash, status in ht:
        nstatus = None
        if status == finished or store.can_load(hash):
            tasks_finished[name] += 1
            nstatus = finished
        else:
            can_run = True
            if status != ready:
                for dep in deps.get(i, []):
                    _, _, dhash, dstatus = ht[dep]
                    if dstatus != finished and not store.can_load(dhash):
                        can_run = False
                        break
            if can_run:
                lock = store.getlock(hash)
                if lock.is_locked():
                    tasks_running[name] += 1
                    nstatus = running
                else:
                    tasks_ready[name] += 1
                    nstatus = ready
            else:
                tasks_waiting[name] += 1
                nstatus = waiting
        assert nstatus is not None, 'update_status: nstatus not assigned'
        if status != nstatus:
            dirty[i] = nstatus
    return tasks_waiting, tasks_ready, tasks_running, tasks_finished, dirty


def _print_status(options, waiting, ready, running, finished):
    if options.short:
        n_ready = sum(ready.values())
        n_running = sum(running.values())
        n_waiting = sum(waiting.values())
        n_finished = sum(finished.values())
        if not n_waiting and not n_running and not n_ready:
            options.print_out('All finished ({0} tasks).'.format(n_finished))
        elif not n_running:
            options.print_out('{0} tasks to be run, {1} finished, (none running).'.format(n_waiting + n_ready, n_finished))
        else:
            options.print_out('{0} tasks to be run, {1} finished, ({2} running).'.format(n_waiting + n_ready, n_finished, n_running))
    else:
        print_task_summary_table(options, [
                                ("Waiting", waiting),
                                ("Ready", ready),
                                ("Finished", finished),
                                ("Running", running)])


def _clear_cache(options):
    from os import unlink
    try:
        unlink(options.status_cache_file)
    except:
        pass


def _status_cached(options):
    create, update = list(range(2))
    try:
        with _open_connection(options) as connection:
            ht, deps, rdeps = retrieve_sqlite3(connection)
        store = backends.select(options.jugdir)
        mode = update
    except:
        store, ht, deps, rdeps = load_jugfile(options)
        mode = create

    tw, tre, tru, tf, dirty = update_status(store, ht, deps, rdeps)
    _print_status(options, tw, tre, tru, tf)
    if mode == update:
        with _open_connection(options) as connection:
            save_dirty3(connection, dirty)
    else:
        for k in dirty:
            _, name, hash, _ = ht[k]
            ht[k] = (k, name, hash, dirty[k])
        with _open_connection(options) as connection:
            create_sqlite3(connection, ht, deps, rdeps)
    return sum(tf.values())


def _status_nocache(options):
    store, _ = jug.init(options.jugfile, options.jugdir)
    Task.store = memoize_store(store, list_base=True)

    tasks_waiting = defaultdict(int)
    tasks_ready = defaultdict(int)
    tasks_running = defaultdict(int)
    tasks_finished = defaultdict(int)
    for t in task.alltasks:
        if t.can_load():
            tasks_finished[t.name] += 1
        elif t.can_run():
            if t.is_locked():
                tasks_running[t.name] += 1
            else:
                tasks_ready[t.name] += 1
        else:
            tasks_waiting[t.name] += 1
    _print_status(options, tasks_waiting, tasks_ready, tasks_running, tasks_finished)
    return sum(tasks_finished.values())


class StatusCommand(SubCommand):
    '''Print status

    status(options)

    Implements the status command.

    Parameters
    ----------
    options : jug options
    '''
    name = "status"

    def run(self, options, *args, **kwargs):
        if options.status_cache:
            try:
                import sqlite3
            except ImportError:
                from sys import stderr
                stderr.write('Cached status relies on sqlite3. Falling back to non-cached version')
                options.status_cache = False
                return status(options)
            if options.status_cache_clear:
                return _clear_cache(options)
            return _status_cached(options)
        else:
            return _status_nocache(options)

    def parse(self, parser):
        defaults = self.parse_defaults()

        parser.add_argument('--cache',
                            action='store_const', const=True,
                            dest='status_cache',
                            help='Use a cache for faster status [does not update after jugfile changes, though]')

        parser.add_argument('--cache-file',
                            action='store', metavar="CACHE_FILE",
                            dest='status_cache_file',
                            help=('Name of file to use for status cache. Use with status --cache. '
                                  '(Default: {status_cache_file}'.format(**defaults)))
        parser.add_argument('--clear',
                            action='store_const', const=True,
                            dest='status_cache_clear',
                            help='Use with status --cache. Removes the cache file')

    def parse_defaults(self):
        return {
            "status_cache": False,
            "status_cache_clear": False,
            "status_cache_file": ".jugstatus.sqlite3",
        }


status = StatusCommand()



# -*- coding: utf-8 -*-
# Copyright (C) 2008-2016, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.


from six.moves import cPickle as pickle
from six import BytesIO
import six
import zlib

__all__ = ['encode', 'decode', 'encode_to', 'decode_from']

def encode(object):
    '''
    s = encode(object)

    Return a string (byte-array) representation of object.

    Parameters
    ----------
      object : Any thing that is pickle()able

    Returns
    -------
      s : string (byte array).

    See
    ---
      `decode`
    '''
    output = BytesIO()
    encode_to(object, output)
    return output.getvalue()

def encode_to(object, stream):
    '''
    encode_to(object, stream)

    Encodes the object into the stream ``stream``

    Parameters
    ----------
    object : Any object
    stream : file-like object
    '''
    if object is None:
        return
    prefix = six.b('P')
    write = lambda f,a: pickle.dump(f, a, protocol=pickle.HIGHEST_PROTOCOL)
    try:
        import numpy as np
        if type(object) == np.ndarray:
            prefix = six.b('N')
            write = (lambda f,a: np.save(a,f))
    except ImportError:
        pass
    stream = compress_stream(stream)
    stream.write(prefix)
    write(object, stream)
    stream.flush()

class compress_stream(object):
    def __init__(self, stream):
        self.stream = stream
        self.C = zlib.compressobj()

    def write(self, s):
        MAX_BLOCK = 2000000000
        if len(s) > MAX_BLOCK:
            self.stream.write(self.C.compress(s[:MAX_BLOCK]))
            self.write(s[MAX_BLOCK:])
        else:
            self.stream.write(self.C.compress(s))

    def flush(self):
        self.stream.write(self.C.flush())
        self.stream.flush()

class decompress_stream(object):
    def __init__(self, stream, block=8192):
        self.stream = stream
        self.D = zlib.decompressobj()
        self.block = block
        self.lastread = six.b('')
        self.queue = six.b('')

    def read(self, nbytes):
        if len(self.queue) >= nbytes:
            res = self.queue[:nbytes]
            self.queue = self.queue[nbytes:]
            self.lastread = res
            return res
        res = self.queue
        self.queue = six.b('')

        if self.D.unconsumed_tail:
            res += self.D.decompress(self.D.unconsumed_tail, nbytes - len(res))
        while len(res) < nbytes:
            buf = self.stream.read(self.block)
            if not buf:
                res += self.D.flush()
                break
            res += self.D.decompress(buf, nbytes - len(res))
        self.lastread = res
        return res

    def seek(self, offset, whence):
        if whence != 1:
            raise NotImplementedError
        while offset > 0:
            nbytes = min(offset, self.block)
            self.read(nbytes)
            offset -= nbytes
        if offset < 0:
            if offset > len(self.lastread):
                raise ValueError('seek too far')
            skip = len(self.lastread) + offset
            self.queue = self.lastread[skip:]

    def readline(self):
        import six
        qi = self.queue.find(six.b('\n'))
        if qi >= 0:
            qi += 1
            res = self.queue[:qi]
            self.queue = self.queue[qi:]
            return res
        line = six.b('')
        while True:
            block = self.read(self.block)
            if not block:
                return line
            ln = block.find(six.b('\n'))
            if ln == -1:
                line += block
            else:
                ln += 1
                line += block[:ln]
                self.seek(ln-len(block), 1)
                return line

def decode(s):
    '''
    object = decode(s)

    Reverses `encode`.

    Parameters
    ----------
      s : a string representation of the object.

    Returns
    -------
      object : the object
    '''
    return decode_from(BytesIO(s))

def decode_from(stream):
    '''
    object = decode_from(stream)

    Decodes the object from the stream ``stream``

    Parameters
    ----------
    stream : file-like object

    Returns
    -------
    object : decoded object
    '''
    stream = decompress_stream(stream)
    prefix = stream.read(1)
    if not prefix:
        return None
    elif prefix == six.b('P'):
        return pickle.load(stream)
    elif prefix == six.b('N'):
        import numpy as np
        return np.load(stream)
    else:
        raise IOError("jug.backend.decode_from: unknown prefix '%s'" % prefix)



# This tests an important regression:
# adding the module to the module map *before* execfile()ing the jugfile makes
# this not work.

from jug import barrier, Task, value
import jug.mapreduce
import math
from functools import reduce

def double(x):
    val = math.sqrt(2.)*math.sqrt(2.)
    return x*val

two = jug.mapreduce.map(double, list(range(20)))
barrier()
def product(vals):
    import operator
    return reduce(operator.mul, vals)
values = product(value(two))



# Copyright (C) 2014, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
# License: MIT

from .register import jug_hook, register_hook, register_hook_once, reset_all_hooks



# Copyright (C) 2014-2015, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
# License: MIT

_hooks = {}

_known_hooks = frozenset([
        'execute.task-loadable',
        'execute.task-executed1',
        'execute.task-pre-execute',
        'execute.finished_pre_status',
        'execute.finished_post_status',
        ])

def jug_hook(name, args=(), kwargs={}):
    '''Call hook

    Calls ``f(*args, **kwargs)`` for all functions registered with name ``name``.

    Parameters
    ----------
    name : str
        Name
    args, kwargs
        passed to functions

    Returns
    -------
    res : list
        A list with the result of all the functions
    '''
    return [f(*args, **kwargs) for f in _hooks.get(name, [])]

_registered = set([])
def register_hook_once(name, code, f):
    '''Register a hook only once

    Parameters
    ----------
    name : str
        Name of hook
    code : str
        An identifier for this registration. The second time this function is
        called with the same (name, code) pair, no action will be taken.
    f : function
        Function to call
        Note that this must be a function, you cannot use it as a decorator

    See Also
    --------
    register_hook
    '''
    if (name,code) not in _registered:
        _registered.add( (name, code) )
        return register_hook(name, f=f)
    elif f is None:
        raise NotImplementedError("jug.register_hook_once: function argument must be used")


def register_hook(name, f=None):
    '''Register a hook

    Known hooks

    execute.task-executed1(Task)
        A single task has been executed. This hook can call ``sys.exit`` to
        exit the jug process (technically, it needs to raise SystemExit).
        Return values are ignored.

    Parameters
    ----------
    name : str
        Identify the hook name
    f : function
        Function to call
    '''
    if name not in _known_hooks:
        raise ValueError('jug.register_hook: {} is not a known hook name (Known are {}.)'.format(name, list(_known_hooks)))
    if f is None:
        from functools import partial
        return partial(register_hook, name)
    _hooks.setdefault(name, []).append(f)

def reset_all_hooks():
    '''Reset all hooks

    This is a destructive actions, which cannot be undone'''
    _registered.clear()
    _hooks.clear()



# -*- coding: utf-8 -*-
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
# Copyright (C) 2008-2016, Luis Pedro Coelho <luis@luispedro.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

'''
============================================
JUG: Coarse Level Parallelisation for Python
============================================

The main use of jug is from the command line::

    jug status jugfile.py
    jug execute jugfile.py

Where ``jugfile.py`` is a Python script using the ``jug`` library.
'''

from .task import TaskGenerator, Task, Tasklet, value, CachedFunction, iteratetask
from .compound import CompoundTaskGenerator, CompoundTask
from .barrier import barrier, bvalue
from .options import set_jugdir

from .jug import init, is_jug_running
from .backends import file_store, dict_store, redis_store

from .jug_version import __version__

__all__ = [
    'Task',
    'Tasklet',
    'TaskGenerator',
    'iteratetask',
    'value',
    'CachedFunction',
    'CompoundTask',
    'CompoundTaskGenerator',
    'barrier',
    'bvalue',

    'set_jugdir',

    'init',
    'is_jug_running',
    'file_store',
    'dict_store',
    'redis_store',
    ]




#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2015, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from collections import defaultdict
import logging
import sys

from .. import task
from ..hooks import jug_hook, register_hook, register_hook_once
from ..io import print_task_summary_table
from ..jug import init
from . import SubCommand, maybe_print_citation_info


__all__ = [
    'execute'
]


def _sigterm(_, __):
    sys.exit(1)


def _log_loadable(t):
    logging.info('Loadable {0}...'.format(t.name))


class TaskStats(object):
    def __init__(self):
        self.loaded = defaultdict(int)
        self.executed = defaultdict(int)
        register_hook('execute.task-loadable', self.loadable)
        register_hook('execute.task-executed1', self.executed1)

    def loadable(self, t):
        self.loaded[t.name] += 1

    def executed1(self, t):
        self.executed[t.name] += 1


class ExecuteCommand(SubCommand):
    '''Execute tasks

    execute(options)

    Implement 'execute' command
    '''
    name = "execute"

    def run(self, options, *args, **kwargs):
        from signal import signal, SIGTERM
        from ..jug import execution_loop

        signal(SIGTERM, _sigterm)
        tasks = task.alltasks
        tstats = TaskStats()
        store = None
        register_hook_once('execute.task-loadable', '_log_loadable', _log_loadable)

        nr_wait_cycles = int(options.execute_nr_wait_cycles)
        noprogress = 0
        failures = False
        while noprogress < nr_wait_cycles:
            del tasks[:]
            store, jugspace = init(options.jugfile, options.jugdir, store=store)
            if options.debug:
                for t in tasks:
                    # Trigger hash computation:
                    t.hash()

            previous = sum(tstats.executed.values())
            failures = execution_loop(tasks, options) or failures
            after = sum(tstats.executed.values())
            done = not jugspace.get('__jug__hasbarrier__', False)
            if done:
                break
            if after == previous:
                from time import sleep
                noprogress += 1
                sleep(int(options.execute_wait_cycle_time))
            else:
                noprogress = 0
        else:
            logging.info('No tasks can be run!')

        jug_hook('execute.finished_pre_status')
        maybe_print_citation_info(options)
        print_task_summary_table(options, [("Executed", tstats.executed), ("Loaded", tstats.loaded)])
        jug_hook('execute.finished_post_status')

        if failures:
            sys.exit(1)

    def parse(self, parser):
        defaults = self.parse_defaults()

        parser.add_argument('--wait-cycle-time', action='store', dest='execute_wait_cycle_time',
                            metavar='WAIT_CYCLE_TIME', type=int,
                            help=("How long to wait in each cycle (in seconds) "
                                  "(Default: {execute_wait_cycle_time})".format(**defaults)))
        parser.add_argument('--nr-wait-cycles', action='store',
                            dest='execute_nr_wait_cycles',
                            metavar='NR_WAIT_CYCLES', type=int,
                            help=("How many wait cycles to do "
                                  "(Default: {execute_nr_wait_cycles})".format(**defaults)))
        parser.add_argument('--target', action='store', dest='execute_target',
                            metavar='TARGET',
                            help="Restrict tasks to execute based on their name")
        parser.add_argument('--keep-going',
                            action='store_const', const=True,
                            dest='execute_keep_going',
                            help='Continue after errors')

    def parse_defaults(self):
        wait_cycle_time = 12

        default_values = {
            "execute_keep_going": False,
            "execute_target": None,
            "execute_wait_cycle_time": wait_cycle_time,
            "execute_nr_wait_cycles": (30 * 60) // wait_cycle_time,
        }

        return default_values


execute = ExecuteCommand()



from jug.subcommands.webstatus import _format_counts

def test_format_counts():
    assert len(_format_counts({'n': 0}, {'n': 1}, {'n': 2}, {'n':3}))
    assert len(_format_counts(
                    {'n': 0, 'n2': 1 },
                    {'n': 1, 'n2': 0 },
                    {'n': 2, 'n2': 3 },
                    {'n': 3, 'n2': 4 }
                    ))



import inspect
import os

from jug.tests.utils import simple_execute
from jug.tests.task_reset import task_reset
import jug.jug


_jugdir = os.path.abspath(inspect.getfile(inspect.currentframe()))
_jugdir = os.path.join(os.path.dirname(_jugdir), 'jugfiles')


@task_reset
def test_w_barrier():
    jugfile = os.path.join(_jugdir, 'custom_hash_function.py')
    store, space = jug.jug.init(jugfile, 'dict_store')
    simple_execute()
    assert space['hash_called']



# -*- coding: utf-8 -*-
# Copyright (C) 2009-2017, Luis Pedro Coelho <luis@luispedro.org>
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.


import os
import re
from functools import partial

from .task import Task, TaskGenerator, Tasklet, value

__all__ = [
    'timed_path',
    'identity',
    'jug_execute',
    'CustomHash',
    'sync_move',
    ]

def _identity(x):
    return x

def identity(x):
    '''
    x = identity(x)

    `identity` implements the identity function as a Task
    (i.e., ``value(identity(x)) == x``)

    This seems pointless, but if ``x`` is, for example, a very large list, then
    using this function might speed up some computations. Consider::

        large = list(range(100000))
        large = jug.utils.identity(large)
        for i in range(100):
            Task(process, large, i)

    This way the list ``large`` is going to get hashed just once. Without the
    call to ``jug.utils.identity``, it would get hashed at each loop iteration.

    https://jug.readthedocs.io/en/latest/utilities.html#identity

    Parameters
    ----------
    x : any object

    Returns
    -------
    x : x
    '''
    if isinstance(x, (Task, Tasklet)):
        return x
    t = Task(_identity, x)
    t.name = 'identity'
    return t

class CustomHash(object):
    '''
    value = CustomHash(obj, hash_function)

    Set a custom hash function

    This is an advanced feature and you can shoot yourself in the foot with it.
    Make sure you know what you are doing. In particular, hash_function should
    be a strong hash: ``hash_function(obj0) == hash_function(obj1)`` is taken
    to imply that ``obj0 == obj1``


    You can use the helpers in the ``jug.hash`` module (in particular
    ``hash_one``) to help you. The implementation of ``timed_path`` is a good
    example of how to use a CustomHash::

        def hash_with_mtime_size(path):
            from .hash import hash_one
            st = os.stat_result(os.stat(path))
            mtime = st.st_mtime
            size = st.st_size
            return hash_one((path, mtime, size))

        def timed_path(path):
            return CustomHash(path, hash_with_mtime_size)

    The ``path`` object (a string or bytes) is wrapped with a hashing function
    which checks the file value.

    Parameters
    ----------
    obj : any object
    hash_function : function
        This should take your object and return a str
    '''
    def __init__(self, obj, hash_function):
        self.obj = obj
        self.hash_function = hash_function

    def __jug_hash__(self):
        return self.hash_function(self.obj)

    def __jug_value__(self):
        return value(self.obj)

def hash_with_mtime_size(path):
    '''hvalue = hash_with_mtime_size(path)

    Computes a hash that depends on the mtime and size of the file ``path``.

    Parameters
    ----------
    path : filepath

    Returns
    -------
    hvalue : bytes
        A hashed version
    '''
    from .hash import hash_one
    st = os.stat_result(os.stat(path))
    mtime = st.st_mtime
    size = st.st_size
    return hash_one((path, mtime, size))


def timed_path(path):
    '''path = timed_path(path)

    Returns a Task object that simply returns `path` with the exception that it
    uses the paths mtime (modification time) and the file size in the hash.
    Thus, if the file is touched or changes size, this triggers an invalidation
    of the results (which propagates to all dependent tasks).

    Parameters
    ----------
    ipath : str
        A filesystem path

    Returns
    -------
    opath : str
        A task equivalent to ``(lambda: ipath)``.
    '''
    return CustomHash(path, hash_with_mtime_size)


def prepare_task_matcher(pattern):
    if re.match(r'/.*/', pattern):
        # Looks like a regular expression
        regex = re.compile(pattern.strip('/'))

    elif '.' in pattern:
        # Looks like a full task name
        regex = pattern.replace('.', '\\.')

    else:
        # A bare function name perhaps?
        regex = re.compile(r'\.' + pattern)

    return partial(re.search, regex)


@TaskGenerator
def jug_execute(args, check_exit=True, run_after=None, return_value=None):
    '''jug_execute(args, check_exit=True, run_after=None, return_value=None)

    Wrapper around ``subprocess.call()``

    The ``run_after`` and ``return_value`` arguments are used only for setting
    dependencies.

    Examples
    --------

    ::

        create_input_tmp = jug_execute(['cp', 'input', 'input.tmp'])
        jug_execute(['wc', '-l', 'input.tmp'], run_after=create_input_tmp)

        # Use the return_value argument:

        oname = 'input.tmp'
        oname = jug_execute(['cp', 'input', oname], return_value=oname)
        jug_execute(['wc', '-l', oname]) # <- Now depends on previous task


    Parameters
    ----------
    args : list of str
    check_exit : boolean, optional
        If true (default), then a non-zero exit results in an exception
    run_after : any, optional
        This is unused by the function, but can be used to order different
        calls to jug_execute
    return_value : any, optional
        A value to return. Used to induce task dependencies
    '''
    import subprocess
    ret = subprocess.call(args)
    if check_exit and ret != 0:
        raise SystemError("Error in system")
    return return_value

def sync_move(src, dst):
    '''Sync the file and move it

    This ensures that the move is truly atomic

    Parameters
    ----------
    src : filename
        Source file
    dst: filename
        Destination file
    '''
    from jug.backends.file_store import fsync_dir
    from os import rename
    fsync_dir(src)
    rename(src, dst)



import jug.backends.encode
from six import BytesIO

def test_encode_decode_empty_string():
    # This is a regression test for
    # https://github.com/luispedro/jug/issues/39
    s = BytesIO()
    jug.backends.encode.encode_to('', s)
    val = jug.backends.encode.decode_from(BytesIO(s.getvalue()))
    assert val == ''



#-*- coding: utf-8 -*-
# Copyright (C) 2009-2014, Luis Pedro Coelho <luis@luispedro.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

'''
dict_store: an in-memory dictionary.

Does not support multiple processes!
'''

import six
from six.moves import cPickle as pickle
from collections import defaultdict

from .base import base_store, base_lock

def _gen_key(key, name):
    if type(name) != six.text_type:
        name = six.text_type(name, 'utf-8')
    return '{0}:{1}'.format(key, name).encode('utf-8')

def _resultname(name):
    return _gen_key('result', name)

def _lockname(name):
    return _gen_key('lock', name)



class dict_store(base_store):
    def __init__(self, backend=None):
        '''
        dict_store(backend=None)

        Parameters
        ----------
        backend : str, optional
                  filename to load/save to
        '''
        if backend is not None:
            try:
                self.store = pickle.load(open(backend))
            except IOError:
                self.store = {}
        else:
            self.store = {}
        self.backend = backend
        self.counts = defaultdict(int)

    def dump(self, object, name):
        '''
        self.dump(object, name)
        '''
        self.store[_resultname(name)] = pickle.dumps(object)
        self.counts[_gen_key('dump:',name)] += 1


    def can_load(self, name):
        '''
        can = can_load(name)
        '''
        self.counts[_gen_key('exists', name)] += 1
        return _resultname(name) in self.store


    def load(self, name):
        '''
        obj = load(name)

        Loads the objects. Equivalent to pickle.load(), but a bit smarter at times.
        '''
        self.counts[_gen_key('load',name)] += 1
        return pickle.loads(self.store[_resultname(name)])


    def remove(self, name):
        '''
        was_removed = remove(name)

        Remove the entry associated with name.

        Returns whether any entry was actually removed.
        '''
        self.counts[_gen_key('del',name)] += 1
        if self.can_load(name):
            self.counts[_gen_key('true-del',name)] += 1
            del self.store[_resultname(name)]


    def cleanup(self, active, keeplocks=False):
        '''
        cleanup()

        Implement 'cleanup' command
        '''
        existing = set(self.store.keys())
        for act in active:
            try:
                existing.remove(_resultname(act))
            except KeyError:
                pass

        if keeplocks:
            for lock in self.listlocks():
                try:
                    existing.remove(_lockname(lock))
                except ValueError:
                    pass

        cleaned = len(existing)
        for superflous in existing:
            del self.store[superflous]

        return cleaned

    def remove_locks(self):
        '''
        removed = store.remove_locks()

        Remove all locks

        Returns
        -------
        removed : int
            Number of locks removed
        '''
        removed = 0
                # we need a copy of the keys because we change it inside
                # iteration:
        for k in list(self.store.keys()):
            if k.startswith(six.b('lock:')):
                del self.store[k]
                removed += 1
        return removed

    def list(self):
        '''
        for key in store.list():
            ...

        Iterates over all the keys in the store
        '''
        for k in self.store.keys():
            if k.startswith(six.b('result:')):
                yield k[len('result:'):]

    def listlocks(self):
        '''
        for key in store.listlocks():
            ...

        Iterates over all the keys in the store
        '''
        for k in self.store.keys():
            if k.startswith(six.b('lock:')):
                yield k[len('lock:'):]


    def getlock(self, name):
        return dict_lock(self.store, self.counts, name)

    def close(self):
        if self.backend is not None:
            pickle.dump(self.store, open(self.backend, 'w'))
            self.backend = None
    __del__ = close


_NOT_LOCKED, _LOCKED = 0,1
class dict_lock(base_lock):
    '''
    dict_lock

    Functions:
    ----------

    get()
        acquire the lock
    release()
        release the lock
    is_locked()
        check lock state
    '''

    def __init__(self, store, counts, name):
        self.name = _lockname(name)
        self.store = store
        self.counts = counts

    def get(self):
        '''
        lock.get()
        '''

        self.counts[_lockname(self.name)] += 1

        previous = self.store.get(self.name, _NOT_LOCKED)
        self.store[self.name] = _LOCKED
        return previous == _NOT_LOCKED


    def release(self):
        '''
        lock.release()

        Removes lock
        '''
        self.counts[_gen_key('unlock', self.name)] += 1
        del self.store[self.name]

    def is_locked(self):
        '''
        locked = lock.is_locked()
        '''
        self.counts[_gen_key('islock', self.name)] += 1
        return (self.store.get(self.name, _NOT_LOCKED) == _LOCKED)

# vim: set ts=4 sts=4 sw=4 expandtab smartindent:



# -*- coding: utf-8 -*-
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
# Copyright (C) 2010-2017, Luis Pedro Coelho <luis@luispedro.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

'''
See more complete documentation at

https://jug.readthedocs.io/en/latest/barrier.html
'''

class BarrierError(Exception):
    '''
    Used to represent a barrier() violation
    '''
    pass

def barrier():
    '''
    barrier()

    In a jug file, it assures that all tasks defined up to now have been
    completed. If not, parsing will (temporarily) stop at that point.

    This ensures that, after calling ``barrier()`` you are free to call
    ``value()`` to get any needed results.

    See Also
    --------
    bvalue : function
        Restricted version of this function. Often faster
    '''
    # The reason to import here instead of at module level is that if some
    # other code does
    # jug.task.alltasks = []
    # we would still be referring to the old version
    from .task import alltasks
    for t in reversed(alltasks):
        if not t.can_load():
            raise BarrierError


def bvalue(t):
    '''
    value = bvalue(t)

    Named after ``barrier``+``value``, ``value = bvalue(t)`` is similar to::

        barrier()
        value = value(t)

    except that it only checks that `t` is complete (and not all tasks) and
    thus can be much faster than a full ``barrier()`` call.

    Thus, ``bvalue`` stops interpreting the Jugfile if its argument has not run
    yet. When it has run, then it returns its value.

    Example
    -------

    A typical use case is the following:

    1. split an input file into blocks of 1,000 lines
    2. process each of these blocks independently.

    You can use the following function to split the input::

        @TaskGenerator
        def split_input_file(ifile):
            splits = []
            output = None
            with open(ifile) as ifile:
                index = 0
                for i,line in enumerate(ifile):
                    if i % 1000 == 0:
                        if output is not None:
                            output.close()
                        ofile = 'splits.{}.txt'.format(index)
                        output = open(ofile, 'wt')
                        index += 1
                        splits.append(ofile)
                    output.write(line)
            if output is not None:
                output.close()
            return splits

        blocks = split_input_file('input.txt')

    And now, use ``bvalue`` to continue processing:


        for b in bvalue(blocks):
            process(b)

    The use of a barrier-type construct (such as ``bvalue``) is necessary for
    this case because you only know how many blocks you have **after running
    part of the pipeline**.


    See Also
    --------
    barrier : Checks that **all** tasks have results available.
    '''
    from .task import value
    try:
        return value(t)
    except:
        raise BarrierError

